! ********************************************************************
! * GOP - Graphic Optimization Program for EHMO ICONC                *
! *                                                                  *
! * Copyright (C) 1997 by Gion Calzaferri                            *
! *                                                                  *
! * This program is free software; you can redistribute it and/or    *
! * modify it under the terms of the GNU General Public License as   *
! * published by the Free Software Foundation; Version 2 of the      *
! * License.                                                         *
! *                                                                  *
! * This program is distributed in the hope that it will be useful,  *
! * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
! * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU *
! * General Public License for more details.                         *
! *                                                                  *
! * You should have received a copy of the GNU General Public        *
! * License along with this program; if not, write to the Free       *
! * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,   *
! * MA 02111-1307 USA                                                *
! ********************************************************************


!------------------------------------------------------------------------------
! PROJECT : GOP                                            by Roman Calzaferri
!------------------------------------------------------------------------------
! FILENAME      : GENFILE_TYPE.f90
! AUTHOR        : Roman Calzaferri
! DATE          : 22. August 1997
! LAST MODIFIED : 1. October 1997
!------------------------------------------------------------------------------
! CONTENTS : 'Classes' GENFILE, VECTOR and VARIATION
!            The classes are implemented as derived types plus member proce-
!            dures. The first argument of the procedure is always the object
!            that is to be accessed.
!            Description:
!            'Class' GENFILE represents a .gen file and allows some editing,
!            as of now of vectors and variations.
!            'Classes' VECTOR and VARIATION represent a vector or a variation, 
!            respectively, a .gen file contains. Their contents can only be 
!            accessed through member routines.
!------------------------------------------------------------------------------
! SPECIALS : Any one calling INIT on a genfile should call CLEANUP on it when 
!            it is no longer used to ensure memory allocated is freed.
!            I would have used allocatable arrays instead of pointers, but
!            the statement 'ALLOCATABLE' is not understood by Microsoft Fortran 
!------------------------------------------------------------------------------
! ASSUMES : - no line in a .gen file is longer than 80 characters
!------------------------------------------------------------------------------
! USED BY  : PROGRAM GOP,     file GOP_MAIN.f90
!            ENERGY_FUNCTION, file ENERGY_FUNCTION_S.f90
!            SEARCH_EMIN,     file SEARCH_EMIN_S.f90
!------------------------------------------------------------------------------
! USES     : -
!------------------------------------------------------------------------------


MODULE GENFILE_TYPE
   IMPLICIT NONE

   INTEGER :: NUMBER_OF_GENFILES = 0

   !! TYPE VECTOR
   TYPE VECTOR
      PRIVATE
         INTEGER              :: FROM_ATOM, TO_ATOM
         DOUBLE PRECISION     :: DISTANCE, BOND_ANGLE, DIHEDRAL_ANGLE
   END TYPE VECTOR

   !! TYPE VARIATION
   TYPE VARIATION
      PRIVATE
         INTEGER               :: NUMBER_OF_VARS, VARY_VECTOR, TYPE_OF_VAR
         DOUBLE PRECISION      :: INCREMENT
         LOGICAL               :: COUPLING
         TYPE(VARIATION), POINTER :: COUPLED_VAR
   END TYPE VARIATION

   !! TYPE GENFILE
   TYPE GENFILE
      PRIVATE
         INTEGER                                :: VECTOR_COUNT, VARIATION_COUNT
         INTEGER                                :: ATOM_COUNT
         CHARACTER(LEN=50)                      :: GENFILE_NAME, TEMPFILE_NAME
         TYPE(VECTOR), DIMENSION(:), POINTER    :: VECTORS
         TYPE(VARIATION), DIMENSION(:), POINTER :: VARIATIONS
         LOGICAL                                :: INITIALIZED
         ! variables needed to store contents of .gen file. they are as of now
         ! used for no other purpose
         INTEGER           :: CHARGE_IT_ROWS_COUNT, ORBIT_OCC_ROWS_COUNT
         CHARACTER(LEN=80) :: SECOND_LINE, TITLE, ICONC_PARAMS, DELTA
         CHARACTER(LEN=2), DIMENSION(:), POINTER :: ATOM_CARD
         CHARACTER(LEN=80), DIMENSION(:), POINTER :: &
                              CHARGE_IT_ROWS, ORBIT_OCC_ROWS
   END TYPE GENFILE

CONTAINS 
   ! -----------------------------------
   ! PROCEDURES TO ACCESS VECTORS
   ! ----------------------------------- 

   FUNCTION GET_DISTANCE(V)
      TYPE(VECTOR), INTENT(IN) :: V
      DOUBLE PRECISION GET_DISTANCE

      GET_DISTANCE = V%DISTANCE
   END FUNCTION GET_DISTANCE
   ! -----------------------------------

   FUNCTION GET_BOND_ANGLE(V)
      TYPE(VECTOR), INTENT(IN) :: V
      DOUBLE PRECISION GET_BOND_ANGLE

      GET_BOND_ANGLE = V%BOND_ANGLE
   END FUNCTION GET_BOND_ANGLE
   ! -----------------------------------

   FUNCTION GET_DIHEDRAL_ANGLE(V)
      TYPE(VECTOR), INTENT(IN) :: V
      DOUBLE PRECISION GET_DIHEDRAL_ANGLE

      GET_DIHEDRAL_ANGLE = V%DIHEDRAL_ANGLE
   END FUNCTION GET_DIHEDRAL_ANGLE
   ! -----------------------------------

   SUBROUTINE  CHANGE_DISTANCE(V, NEW_DISTANCE)
      TYPE(VECTOR), INTENT(INOUT) :: V
      DOUBLE PRECISION, INTENT(IN)            :: NEW_DISTANCE

      V%DISTANCE = NEW_DISTANCE
   END SUBROUTINE CHANGE_DISTANCE
   ! -----------------------------------

   SUBROUTINE CHANGE_BOND_ANGLE(V, NEW_BOND_ANGLE)
      TYPE(VECTOR), INTENT(INOUT) :: V
      DOUBLE PRECISION, INTENT(IN)            :: NEW_BOND_ANGLE

      V%BOND_ANGLE = NEW_BOND_ANGLE
   END SUBROUTINE CHANGE_BOND_ANGLE
   ! -----------------------------------

   SUBROUTINE CHANGE_DIHEDRAL_ANGLE(V, NEW_DIHED_ANGLE)
      TYPE(VECTOR), INTENT(INOUT) :: V
      DOUBLE PRECISION, INTENT(IN)            :: NEW_DIHED_ANGLE

      V%DIHEDRAL_ANGLE = NEW_DIHED_ANGLE
   END SUBROUTINE CHANGE_DIHEDRAL_ANGLE
   ! -----------------------------------

   SUBROUTINE PRINT_VECTOR_INFO(VS)
      ! prints contents of an array of vectors
      TYPE(VECTOR), DIMENSION(:) :: VS
      INTEGER                    :: N, I

      N = SIZE(VS)

      PRINT '(A,/,A,/,A)', &
            " vector- vector  distance  bond-     dihedral- ", &
            " number                    angle     angle     ", &
            " ------- ------- --------  --------  --------  "


      DO I=1,N
         PRINT '(I4, "    ", 2I4, 3F10.5)', I, VS(I)%FROM_ATOM, VS(I)%TO_ATOM, &
            VS(I)%DISTANCE, VS(I)%BOND_ANGLE, VS(I)%DIHEDRAL_ANGLE
      END DO
   END SUBROUTINE PRINT_VECTOR_INFO
   ! -----------------------------------
   ! -----------------------------------


   ! -----------------------------------
   ! PROCEDURES TO ACCES VARIATIONS
   ! -----------------------------------

   SUBROUTINE CHANGE_NUMBER_OF_VARS(VA, NEW_NUMBER)
      TYPE(VARIATION), INTENT(INOUT) :: VA
      INTEGER, INTENT(IN)            :: NEW_NUMBER

      VA%NUMBER_OF_VARS = NEW_NUMBER
   END SUBROUTINE CHANGE_NUMBER_OF_VARS
   ! -----------------------------------
      
   SUBROUTINE CHANGE_INCREMENT(VA, NEW_INCREMENT)
      TYPE(VARIATION), INTENT(INOUT) :: VA
      DOUBLE PRECISION, INTENT(IN)               :: NEW_INCREMENT

      VA%INCREMENT = NEW_INCREMENT
   END SUBROUTINE CHANGE_INCREMENT
   ! -----------------------------------

   FUNCTION GET_VARY_VECTOR(VA)
      TYPE(VARIATION), INTENT(IN) :: VA
      INTEGER GET_VARY_VECTOR

      GET_VARY_VECTOR = VA%VARY_VECTOR
   END FUNCTION GET_VARY_VECTOR
   ! -----------------------------------
      
   FUNCTION GET_TYPE_OF_VAR(VA)
      TYPE(VARIATION), INTENT(IN) :: VA
      INTEGER GET_TYPE_OF_VAR

      GET_TYPE_OF_VAR = VA%TYPE_OF_VAR
   END FUNCTION GET_TYPE_OF_VAR
   ! -----------------------------------
   ! -----------------------------------


   ! -----------------------------------
   ! PROCEDURES TO ACCESS GENFILE
   ! -----------------------------------

   SUBROUTINE INIT(GF, ACCESS_NAME, ERR)
      ! loads the contents of an existing .gen file and creates a temporary
      ! .gen file
      TYPE(GENFILE), INTENT(INOUT) :: GF
      CHARACTER(LEN=*), INTENT(IN) :: ACCESS_NAME
      INTEGER, OPTIONAL            :: ERR
      INTEGER                      :: I, ERROR_STATE, BUFFER

      
      NUMBER_OF_GENFILES = NUMBER_OF_GENFILES + 1
      WRITE(GF%TEMPFILE_NAME,'(A, I1, A)') 'TEMP$', &
                    NUMBER_OF_GENFILES, '.GEN'
      GF%GENFILE_NAME  = ACCESS_NAME

      ! read in genfile
      OPEN(UNIT=1, FILE=GF%GENFILE_NAME, STATUS='OLD', & 
                                         ACTION='READ', IOSTAT = ERROR_STATE)
      IF (ERROR_STATE /= 0) THEN
         IF (PRESENT(ERR)) ERR = ERROR_STATE
         RETURN
      END IF

      ! read first+second line (region a)
      READ(1,'(4I4)') GF%VECTOR_COUNT, GF%VARIATION_COUNT, &
                      GF%CHARGE_IT_ROWS_COUNT, GF%ORBIT_OCC_ROWS_COUNT
      READ(1, '(A)') GF%SECOND_LINE
      
      !(regions b,c,d)
      READ(1, '(A)') GF%TITLE
      READ(1, '(A)') GF%ICONC_PARAMS
      READ(1, '(A)') GF%DELTA

      !read the vectors (region e)
      ALLOCATE(GF%VECTORS(GF%VECTOR_COUNT))
      CALL READ_VECTORS


      !read number of atoms from region c
      READ(GF%ICONC_PARAMS, '(2I3)') BUFFER, GF%ATOM_COUNT

      ALLOCATE(GF%ATOM_CARD(GF%ATOM_COUNT))

      !(region f)
      READ(1, '(40A2)') (GF%ATOM_CARD(I), I=1,GF%ATOM_COUNT)


      ALLOCATE(GF%CHARGE_IT_ROWS(GF%CHARGE_IT_ROWS_COUNT))
      ALLOCATE(GF%ORBIT_OCC_ROWS(GF%ORBIT_OCC_ROWS_COUNT))

      !(region g+h)
      DO I = 1, GF%CHARGE_IT_ROWS_COUNT
         READ(1, '(A)') GF%CHARGE_IT_ROWS(I)
      END DO

      !(region i,j,k,l)
      DO I = 1, GF%ORBIT_OCC_ROWS_COUNT
         READ(1, '(A)') GF%ORBIT_OCC_ROWS(I)
      END DO

      ! read variations (region m)
      ALLOCATE(GF%VARIATIONS(GF%VARIATION_COUNT))
      CALL READ_VARIATIONS

      CLOSE(1)


      GF%INITIALIZED = .TRUE.
      !now write the temporary file
      GF%TITLE = 'TEMPORARY FILE'

      CALL WRITE_TEMPFILE(GF, ERROR_STATE)
      IF (ERROR_STATE /= 0) THEN
         IF (PRESENT(ERR)) ERR = ERROR_STATE + 10
         RETURN
      END IF

      !----------------------------
      ! internal procedures of init
      CONTAINS  
         SUBROUTINE READ_VECTORS
            ! reads in the vectors of the .gen file
            INTEGER I

               DO I = 1, GF%VECTOR_COUNT
                  READ(1,'(I3,I4, 3F10.5)')   GF%VECTORS(I)%FROM_ATOM, &
                     GF%VECTORS(I)%TO_ATOM, GF%VECTORS(I)%DISTANCE, &
                     GF%VECTORS(I)%BOND_ANGLE, GF%VECTORS(I)%DIHEDRAL_ANGLE
               END DO
               
         END SUBROUTINE READ_VECTORS
         ! -----------------------------------

         SUBROUTINE READ_VARIATIONS
            ! reads in the variations of the .gen file
            INTEGER I
            CHARACTER(LEN=2)  ::COUPLE_BUFFER

            DO I = 1, GF%VARIATION_COUNT
               READ(1, '(I3,A2,2I3,F10.6)')                         &
                    GF%VARIATIONS(I)%NUMBER_OF_VARS, COUPLE_BUFFER, &
                    GF%VARIATIONS(I)%VARY_VECTOR,                   &
                    GF%VARIATIONS(I)%TYPE_OF_VAR,                   &
                    GF%VARIATIONS(I)%INCREMENT
               IF (COUPLE_BUFFER(1:1) == '*' .OR. COUPLE_BUFFER(2:2) == '*') THEN
                  GF%VARIATIONS(I)%COUPLING = .TRUE.
                  GF%VARIATIONS(I)%COUPLED_VAR => GF%VARIATIONS(I-1)
               ELSE
                  GF%VARIATIONS(I)%COUPLING = .FALSE.
                  NULLIFY(GF%VARIATIONS(I)%COUPLED_VAR)
               END IF
            END DO
         END SUBROUTINE READ_VARIATIONS
         ! -----------------------------------

   END SUBROUTINE INIT
   ! -----------------------------------

   SUBROUTINE WRITE_TEMPFILE(GF, ERR)
      ! writes the temporary file to the disk.
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(IN) :: GF
      INTEGER, OPTIONAL         :: ERR
      INTEGER I, ERROR_STATE

      OPEN(UNIT=1, FILE=GF%TEMPFILE_NAME, STATUS='REPLACE', &
                                          IOSTAT = ERROR_STATE)
      IF (ERROR_STATE /= 0) THEN
         IF (PRESENT(ERR)) ERR = ERROR_STATE
         RETURN
      END IF

      !(region a,b,c,d)
      WRITE(1, '(4I4)')  GF%VECTOR_COUNT, GF%VARIATION_COUNT, &
                         GF%CHARGE_IT_ROWS_COUNT, GF%ORBIT_OCC_ROWS_COUNT

      WRITE(1, '(A)') GF%SECOND_LINE
      WRITE(1, '(A)') GF%TITLE
      WRITE(1, '(A)') GF%ICONC_PARAMS
      WRITE(1, '(A)') GF%DELTA

      !(region e)
      CALL WRITE_VECTORS

      !(region f)
      WRITE(1, '(40A2)') (GF%ATOM_CARD(I), I=1,GF%ATOM_COUNT)

      !(region g,h)
      DO I = 1, GF%CHARGE_IT_ROWS_COUNT
         WRITE(1, '(A)') GF%CHARGE_IT_ROWS(I)
      END DO

      !(region i,j,k,l)
      DO I = 1, GF%ORBIT_OCC_ROWS_COUNT
         WRITE(1, '(A)') GF%ORBIT_OCC_ROWS(I)
      END DO

      !(region l)
      CALL WRITE_VARIATIONS

      WRITE(1, '(/)')

      CLOSE(1)

      !--------------------------------------
      ! internal procedures of write_tempfile
      CONTAINS 
         SUBROUTINE WRITE_VECTORS
            INTEGER I

               DO I = 1, GF%VECTOR_COUNT
                  WRITE(1,'(I3, I4, 3F10.5)')   GF%VECTORS(I)%FROM_ATOM, &
                     GF%VECTORS(I)%TO_ATOM, GF%VECTORS(I)%DISTANCE, &
                     GF%VECTORS(I)%BOND_ANGLE, GF%VECTORS(I)%DIHEDRAL_ANGLE
               END DO
               
         END SUBROUTINE WRITE_VECTORS
         ! -----------------------------------

         SUBROUTINE WRITE_VARIATIONS
            INTEGER I
            CHARACTER(LEN=2)  ::COUPLE_BUFFER

            DO I = 1, GF%VARIATION_COUNT
               IF (GF%VARIATIONS(I)%COUPLING) THEN
                  COUPLE_BUFFER = ' *'
               ELSE
                  COUPLE_BUFFER = '  '
               END IF
               WRITE(1, '(I3,A2,2I3,F10.6)')                         &
                    GF%VARIATIONS(I)%NUMBER_OF_VARS, COUPLE_BUFFER, &
                    GF%VARIATIONS(I)%VARY_VECTOR,                   &
                    GF%VARIATIONS(I)%TYPE_OF_VAR,                   &
                    GF%VARIATIONS(I)%INCREMENT
            END DO
         END SUBROUTINE WRITE_VARIATIONS
         ! -----------------------------------

   END SUBROUTINE WRITE_TEMPFILE
   ! -----------------------------------

   SUBROUTINE MINIMIZE_OUTPUT(GF)
      !minimizes the output iconc will make by changing the flags in region c
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(INOUT) :: GF
      INTEGER I

      DO I = 45, 65
         GF%ICONC_PARAMS(I:I) = 'T'
      END DO
   END SUBROUTINE MINIMIZE_OUTPUT
   ! -----------------------------------

   FUNCTION GET_VARIATIONS(GF)
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(IN) :: GF
      TYPE(VARIATION), DIMENSION(GF%VARIATION_COUNT) :: GET_VARIATIONS

      GET_VARIATIONS = GF%VARIATIONS
   END FUNCTION GET_VARIATIONS
   ! -----------------------------------

   FUNCTION GET_VECTORS(GF)
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(IN) :: GF
      TYPE(VECTOR), DIMENSION(GF%VECTOR_COUNT) :: GET_VECTORS

      GET_VECTORS = GF%VECTORS
   END FUNCTION GET_VECTORS
   ! -----------------------------------

   FUNCTION GET_VARIATION_COUNT(GF)
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(IN) :: GF
      INTEGER                   :: GET_VARIATION_COUNT
      
      GET_VARIATION_COUNT = GF%VARIATION_COUNT
   END FUNCTION GET_VARIATION_COUNT      
   ! -----------------------------------

   FUNCTION GET_VECTOR_COUNT(GF)
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(IN) :: GF
      INTEGER                   :: GET_VECTOR_COUNT
      
      GET_VECTOR_COUNT = GF%VECTOR_COUNT
   END FUNCTION GET_VECTOR_COUNT      
   ! -----------------------------------

   SUBROUTINE DELETE_VARIATIONS(GF)
      ! sets the variation count to 0 -> no variations will be written to
      ! the temporary .gen file
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(INOUT) :: GF

      GF%VARIATION_COUNT = 0
   END SUBROUTINE DELETE_VARIATIONS
   ! -----------------------------------

   FUNCTION GET_TEMPFILE_NAME(GF)
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(IN) :: GF
      CHARACTER(LEN=50) :: GET_TEMPFILE_NAME

      GET_TEMPFILE_NAME = GF%TEMPFILE_NAME
   END FUNCTION GET_TEMPFILE_NAME
   ! -----------------------------------

   FUNCTION GET_GENFILE_NAME(GF)
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(IN) :: GF
      CHARACTER(LEN=50) :: GET_GENFILE_NAME

      GET_GENFILE_NAME = GF%GENFILE_NAME
   END FUNCTION GET_GENFILE_NAME
   ! -----------------------------------

   SUBROUTINE CHANGE_VECTORS(GF,VECS)
      ! overwrites the vectors in the .gen file. The number of vectors must 
      ! stay the same!
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(INOUT)                         :: GF
      TYPE(VECTOR), DIMENSION(GF%VECTOR_COUNT), INTENT(IN) :: VECS

      GF%VECTORS = VECS
   END SUBROUTINE CHANGE_VECTORS
   ! -----------------------------------

   SUBROUTINE PRINT_VARIATIONS(GF)
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(IN) :: GF
      INTEGER I

      DO I = 1,GF%VARIATION_COUNT
         WRITE(*, '(I4,L4,2I4,F10.5)')               &   
                    GF%VARIATIONS(I)%NUMBER_OF_VARS, &
                    GF%VARIATIONS(I)%COUPLING,        &
                    GF%VARIATIONS(I)%VARY_VECTOR,    &
                    GF%VARIATIONS(I)%TYPE_OF_VAR,    &
                    GF%VARIATIONS(I)%INCREMENT
      END DO
   END SUBROUTINE PRINT_VARIATIONS
   ! -----------------------------------

   SUBROUTINE PRINT_VECTORS(GF)
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(IN) :: GF

      CALL PRINT_VECTOR_INFO(GF%VECTORS)
   END SUBROUTINE PRINT_VECTORS
   ! -----------------------------------

   SUBROUTINE CLEANUP(GF)
      !------------------------------
      ! REQUIRES : INITIALIZED = .TRUE.
      !------------------------------
      TYPE(GENFILE), INTENT(INOUT) :: GF

      DEALLOCATE(GF%VECTORS, GF%VARIATIONS, GF%CHARGE_IT_ROWS, &
                 GF%ORBIT_OCC_ROWS, GF%ATOM_CARD)
      GF%INITIALIZED = .FALSE.
   END SUBROUTINE CLEANUP
   ! -----------------------------------
   ! -----------------------------------
      

END MODULE GENFILE_TYPE   
! -----------------------------------
! end of file
! -----------------------------------

